#include <stdio.h>
#include <math.h>
#include "fftpert.h"

#define NMAX 512

double bekep(DOUBLE ex, DOUBLE ey, DOUBLE x0)
{

/*

     bekep: Bisection method for Elliptic KEPler equation

     Eq.: D - e_X sin D + e_Y cos D = L

     Ref: T. Fukushima, 1996, Astron. J, 112, 2858

     Output Variable: bekep = D (Solution)

     Input Variables: ex = e_X, ey = e_Y, x0 = L


     Re-coded by S. Klioner into C from the original author's fortran code 

*/

/*  Local Arrays */

      static DOUBLE a[NMAX+1],s[NMAX+1],c[NMAX+1];

/* Constants */                               

      static const DOUBLE ONEPI = PI;
      static const DOUBLE EPSLON = BEKEPEPSLON;
      static const DOUBLE TWOPI = PI * 2.e0;
      static const DOUBLE A3 = 1.0/3.0;
      static const DOUBLE A6 = 1.0/6.0;
      static const DOUBLE A30 =1.0/30.0;
      static const DOUBLE A90 = 1.0/90.0;
      static const DOUBLE A620 = 1.0/620.0;

      static const INT NDIM0=7;
      static const INT NDIM1=8;
      static const INT NDIM2=9;
      static const INT MEND=30;

/*     Local Variables */

      INT ixend,ndim,i;
      DOUBLE x,dx,theta,f0,f1,f2,f3,d,d2,sd,dcd,df2,df3,fx,fy,fz,z;

      INT n,ix,m;
      DOUBLE e2;

      static INT first = 1;

/* Preparation of Auxiliary Tables */

      if ( first == 1)
	{
         first = 0;
         ixend = 1;
         i = 1;
         a[1] = ONEPI;
         s[1] = 0.0;
         c[1] = -1.0;
         theta = ONEPI;
         for(n=2; n <= NDIM2; n++)
	   {
            ixend = ixend * 2;
            theta = theta * 0.5;
            for(ix=1; ix <= ixend; ix++)
	      {
               i++;
               a[i] = theta * (2 * ix -1);
               s[i] = sin(a[i]);
               c[i] = cos(a[i]);
	      }
	   }
	}

/*     Selection of Bisection Level (Revised Nov. 11, 1996) */

      e2 = ex * ex + ey * ey;

      if(e2 > 1.0) 
	{
         printf("(bekep) Hyperbolic case. %10.6f %10.6f\n", ex,ey);
         return 0.0;
	} 
      else if(e2 < 0.5)
	{
         ndim = NDIM0;
	}   
      else if(e2 < 0.75)
	{
         ndim = NDIM1;
	}
      else
	{
         ndim = NDIM2;
	}

/* Reduction of 'L' */

      if( x0 < ey )
	{
         x = ey - fmod(ey-x0,TWOPI) + TWOPI;
	}
      else if( x0 > ey + TWOPI )
	{
         x = fmod(x0-ey,TWOPI) + ey;
	}
      else
	{
         x = x0;
	}

      dx = x0 - x;

/* Bisection Search */

      i = 1;
      for(n=2; n<=ndim; n++)
	{
         f0 = (a[i] - (ex * s[i] - ey * c[i]) ) - x;
          if( f0 > 0.0 )
	    {
             i = i * 2;
	    }
          else
	    {
             i = i * 2 + 1;
	    }
	}

/*  Newton Method by using Taylor Expanded Trig. Func. */    

      f2 = ex * s[i] - ey * c[i];
      f3 = ex * c[i] + ey * s[i];
      f1 = 1.0 - f3;
      f0 = ( a[i] - f2 ) - x;
      d = -f0/f1;
      n = 0;
      for(m=1; m<=MEND; m++)
	{
         d2 = d * d * 0.5;
         sd = d*(1.0-d2*(A3-d2*(A30-d2*A620)));
         dcd = d2*(1.0-d2*(A6-d2*A90));
         df2 = -f2*dcd+f3*sd;
         df3 = -f3*dcd-f2*sd;
         fx = f1-df3;
         fy = (f0+d)-df2;
         d = d-fy/fx;
         if(fabs(fy)<EPSLON) 
	   {
            n = 1;         
            break;
	   }
	}
 

/* In case of No convergence */
      
      if( n == 0 )
	{
         z = a[i] + d;
         fz = (z-(ex*sin(z)-ey*cos(z)))-x;
         printf("(bekep) No Conv. "FORDOUBLE" "FORDOUBLE" "FORDOUBLE" "FORDOUBLE"\n", ex,ey,x,fz);
	}

/* Converged */
     
/* Recovery of 'L' Reduction */     

      if( fabs(dx) > 0.0 )
	{
         return a[i]+d+dx;
	}
      else
	{
         return a[i]+d;
	}

}

