/* Bessel function J_n(x) for real x and integer n */

/* Sergei A. Klioner, 1999 */

/*

Re-coded from W.H. Press, S.A. Teukolsky, W.T. Vetterling, B.P. Flannery,
Numerical Recipes in C, Cambridge University Press, 1992 (reprinted 1995,
software version 2.06, pages 234-235)

*/

/* 

   Absolute errors with 

egs-2.91.60 on SGI (IP27) under IRIX 6.5 (64 bit double)
and Maple V.5 on the same computer with 32 digit arithmetics

J_i(e*i):

e=0.1       i<=101       < 2.5e-18  (for i >  101  |J_i(e*i)| < 1.2e-090)
 
e=0.3       i<=300       < 3.0e-18  (for i >  301  |J_i(e*i)| < 1.4e-122)

e=0.5       i<=1001      < 1.4e-17  (for i > 1002  |J_i(e*i)| < 8.0e-199)

e=0.7       i<=1728      < 4.5e-17  (for i > 1729  |J_i(e*i)| < 3.2e-139)

e=0.9       i<=5720      < 5.0e-17  (for i > 5721  |J_i(e*i)| < 1.8e-080)

e=0.999999  i<=10000     < 7.2e-16

For e<1.0 the relative error (precision) is more or less constant:
the error is independent of those of J_0 and J_1.

e=1.0       i<=23000     < 1.11e-15

e=2.0       i<=18700     < 4.5e-16

e=10.0      i<=26200     < 1.5e-14
e=10.0      i<=33800     < 4.3e-14

e=100.0     i<=  2600    < 1.2e-14
e=100.0     i<= 10400    < 3.9e-14
e=100.0     i<= 41900    < 6.3e-14
e=100.0     i<=100000    < 1.7e-13

The errors come probably from the errors of J_0 and J_1.
For e>1.0 the absolute error is getting slowly larger with i.

*/

#include <math.h>
#include "fftpert.h"

#define ACC   300.0
#define BIGNO 1.0e20
#define BIGNI 1.0e-20

extern DOUBLE besselj0 (DOUBLE);
extern DOUBLE besselj1 (DOUBLE);

DOUBLE besseljn(INT n, DOUBLE x)

{
	INT j,jsum,m,an;
	DOUBLE ax,bj,bjm,bjp,sum,tox,ans;
        
/* Bessel J_0(x) */
        if (n==0) 
	  {
           return(besselj0(x));
	  }
/* Bessel J_1(x) or J_-1(x) */
        else if (n==1 || n==-1)
	  {
           return(n*besselj1(x));
	  }
/* Bessel J_n(x), |n|>=2 */
        else
	  {
	   an=abs(n);
           ax=fabs(x);
/* Bessel J_n(0)=0, |n|>=2 */
           if (ax == 0.0) 
	     {
              return(0.0e0);
	     }
/* Bessel J_n(x), |n|>=2 and |x|>|n|: J_(n+1)(x) = 2n/x J_n(x) - J_(n-1)(x) */
           else if (ax > (double) an) 
             {
		tox=2.0/ax;
		bjm=besselj0(ax);
		bj=besselj1(ax);
		for (j=1;j<an;j++) 
                  {
		   bjp=j*tox*bj-bjm;
		   bjm=bj;
		   bj=bjp;
		  }
		ans=bj;
	     } 
/* Bessel J_n(x), |n|>=2 and |x|<|n|: J_(n-1)(x) = 2n/x J_n(x) - J_(n+1)(x) */
           else
              {
		tox=2.0/ax;
		m=2*((an+(int) sqrt(ACC*an))/2);
		jsum=0;
		bjp=ans=sum=0.0;
		bj=1.0;
		for (j=m;j>0;j--) 
                   {
		    bjm=j*tox*bj-bjp;
		    bjp=bj;
		    bj=bjm;
		    if (fabs(bj) > BIGNO) 
                       {
			bj *= BIGNI;
			bjp *= BIGNI;
			ans *= BIGNI;
			sum *= BIGNI;
		       }
		     if (jsum) sum += bj;
		     jsum=!jsum;
		     if (j == an) ans=bjp;
		   }
		sum=2.0*sum-bj;
		ans /= sum;
	      }
	  }
/* Answer with symmetries in signs of n and x */
        j = +1;
        m = an % 2;
        if (n<0 && m==1) j=-j;
        if (x<0 && m==1) j=-j;
	return(j < 0 ? -ans : ans);
}
